/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 03/07/22.
//

#ifndef FGGL_ASSETS_TYPES_HPP
#define FGGL_ASSETS_TYPES_HPP

#include <filesystem>
#include <variant>
#include <functional>

#include "fggl/util/safety.hpp"
#include "fggl/util/guid.hpp"

namespace fggl::assets {
	using AssetType = util::OpaqueName<std::string_view, struct AssetTag>;
	using AssetGUID = std::string;
	using AssetPath = std::filesystem::path;

	using AssetID = util::OpaqueName<uint64_t, struct AssetIDTag>;

	template<unsigned L1, unsigned L2>
	constexpr AssetID make_asset_id(const char (&pack)[L1], const char (&path)[L2]) {
		auto hash = util::hash_fnv1a_64( util::cat( pack, ":", path ).c );
		return AssetID::make( hash );
	}

	template<unsigned L1, unsigned L2, unsigned L3>
	constexpr AssetID make_asset_id(const char (&pack)[L1], const char (&path)[L2], const char (&view)[L3]) {
		auto hash = util::hash_fnv1a_64( util::cat( pack, ":", path, "[", view, "]").c );
		return AssetID::make( hash );
	}

	AssetID make_asset_id_rt(const std::string &pack, const std::string &path, const std::string &view = "");
	AssetID asset_from_user(const std::string &input, const std::string &pack = "core");

	using AssetTypeID = util::OpaqueName<uint64_t, struct AssetTypeTag>;
	constexpr auto INVALID_ASSET_TYPE = AssetTypeID::make(0);

	constexpr AssetTypeID make_asset_type(const char* type) {
		return AssetTypeID::make( util::hash_fnv1a_64(type) );
	}


	struct MemoryBlock {
		std::byte *data;
		std::size_t size;

		template<typename T>
		T* viewAs(std::size_t offset = 0) {
			static_assert( std::is_standard_layout<T>::value );
			return (T*)( data[offset] );
		}
	};

	using AssetRefRaw = std::shared_ptr<void>;

	struct LoaderContext {
		std::string pack;
		std::filesystem::path packRoot;
		std::filesystem::path assetPath;
	};

	class Loader;
	using Checkin = std::function<AssetRefRaw(Loader* loader, const AssetID &, const LoaderContext &, void* userPtr)>;
}

// formatter
template<> struct fmt::formatter<fggl::assets::AssetID> {
	constexpr auto parse(format_parse_context& ctx) -> decltype(ctx.begin()) {
		return ctx.begin();
	}

	template <typename FormatContext>
	auto format(const fggl::assets::AssetID & guid, FormatContext& ctx) const -> decltype(ctx.out()) {
		#ifndef NDEBUG
		return fmt::format_to(ctx.out(), "ASSET[{}]", guidToString( fggl::util::GUID::make( guid.get() ) ));
		#else
		return fmt::format_to(ctx.out(), "ASSET[{}]", guid.get());
		#endif
	}
};

#endif //FGGL_ASSETS_TYPES_HPP
